@@ -3,8 +3,13 @@ require 'mkmf'
 have_header("ruby/st.h")
 have_header("st.h")
 have_func("rb_str_replace", ["ruby.h"])
+have_func("rb_intern_str", ["ruby.h"])
+have_func("rb_sym2str", ["ruby.h"])
+have_func("rb_str_intern", ["ruby.h"])
 
-$CFLAGS << %[ -I.. -Wall -O3 -g -std=c99]
+unless RUBY_PLATFORM.include? 'mswin'
+  $CFLAGS << %[ -I.. -Wall -O3 -g -std=c99]
+end
 #$CFLAGS << %[ -DDISABLE_RMEM]
 #$CFLAGS << %[ -DDISABLE_RMEM_REUSE_INTERNAL_FRAGMENT]
 #$CFLAGS << %[ -DDISABLE_BUFFER_READ_REFERENCE_OPTIMIZE]
