@@ -1,7 +1,7 @@
 /*
  * MessagePack for Ruby
  *
- * Copyright (C) 2008-2012 FURUHASHI Sadayuki
+ * Copyright (C) 2008-2013 Sadayuki Furuhashi
  *
  *    Licensed under the Apache License, Version 2.0 (the "License");
  *    you may not use this file except in compliance with the License.
@@ -28,19 +28,11 @@
 static msgpack_rmem_t s_stack_rmem;
 #endif
 
-#ifdef COMPAT_HAVE_ENCODING  /* see compat.h*/
-static int s_enc_utf8;
-#endif
-
 void msgpack_unpacker_static_init()
 {
 #ifdef UNPACKER_STACK_RMEM
     msgpack_rmem_init(&s_stack_rmem);
 #endif
-
-#ifdef COMPAT_HAVE_ENCODING
-    s_enc_utf8 = rb_utf8_encindex();
-#endif
 }
 
 void msgpack_unpacker_static_destroy()
@@ -50,9 +42,9 @@ void msgpack_unpacker_static_destroy()
 #endif
 }
 
-#define HEAD_BYTE_REQUIRED 0xc6
+#define HEAD_BYTE_REQUIRED 0xc1
 
-void msgpack_unpacker_init(msgpack_unpacker_t* uk)
+void _msgpack_unpacker_init(msgpack_unpacker_t* uk)
 {
     memset(uk, 0, sizeof(msgpack_unpacker_t));
 
@@ -73,7 +65,7 @@ void msgpack_unpacker_init(msgpack_unpacker_t* uk)
     uk->stack_capacity = MSGPACK_UNPACKER_STACK_CAPACITY;
 }
 
-void msgpack_unpacker_destroy(msgpack_unpacker_t* uk)
+void _msgpack_unpacker_destroy(msgpack_unpacker_t* uk)
 {
 #ifdef UNPACKER_STACK_RMEM
     msgpack_rmem_free(&s_stack_rmem, uk->stack);
@@ -101,7 +93,7 @@ void msgpack_unpacker_mark(msgpack_unpacker_t* uk)
     rb_gc_mark(uk->buffer_ref);
 }
 
-void msgpack_unpacker_reset(msgpack_unpacker_t* uk)
+void _msgpack_unpacker_reset(msgpack_unpacker_t* uk)
 {
     msgpack_buffer_clear(UNPACKER_BUFFER_(uk));
 
@@ -149,10 +141,17 @@ static inline int object_complete(msgpack_unpacker_t* uk, VALUE object)
 
 static inline int object_complete_string(msgpack_unpacker_t* uk, VALUE str)
 {
-    // TODO ruby 2.0 has String#b method
 #ifdef COMPAT_HAVE_ENCODING
-    //str_modifiable(str);
-    ENCODING_SET(str, s_enc_utf8);
+    ENCODING_SET(str, msgpack_rb_encindex_utf8);
+#endif
+    return object_complete(uk, str);
+}
+
+static inline int object_complete_binary(msgpack_unpacker_t* uk, VALUE str)
+{
+#ifdef COMPAT_HAVE_ENCODING
+    // TODO ruby 2.0 has String#b method
+    ENCODING_SET(str, msgpack_rb_encindex_ascii8bit);
 #endif
     return object_complete(uk, str);
 }
@@ -248,7 +247,7 @@ static int read_raw_body_cont(msgpack_unpacker_t* uk)
     return PRIMITIVE_OBJECT_COMPLETE;
 }
 
-static inline int read_raw_body_begin(msgpack_unpacker_t* uk)
+static inline int read_raw_body_begin(msgpack_unpacker_t* uk, bool str)
 {
     /* assuming uk->reading_raw == Qnil */
 
@@ -257,9 +256,16 @@ static inline int read_raw_body_begin(msgpack_unpacker_t* uk)
     if(length <= msgpack_buffer_top_readable_size(UNPACKER_BUFFER_(uk))) {
         /* don't use zerocopy for hash keys but get a frozen string directly
          * because rb_hash_aset freezes keys and it causes copying */
-        bool frozen = is_reading_map_key(uk);
-        VALUE string = msgpack_buffer_read_top_as_string(UNPACKER_BUFFER_(uk), length, frozen);
-        object_complete_string(uk, string);
+        bool will_freeze = is_reading_map_key(uk);
+        VALUE string = msgpack_buffer_read_top_as_string(UNPACKER_BUFFER_(uk), length, will_freeze);
+        if(str == true) {
+            object_complete_string(uk, string);
+        } else {
+            object_complete_binary(uk, string);
+        }
+        if(will_freeze) {
+            rb_obj_freeze(string);
+        }
         uk->reading_raw_remaining = 0;
         return PRIMITIVE_OBJECT_COMPLETE;
     }
@@ -285,14 +291,14 @@ static int read_primitive(msgpack_unpacker_t* uk)
     SWITCH_RANGE(b, 0xe0, 0xff)  // Negative Fixnum
         return object_complete(uk, INT2NUM((int8_t)b));
 
-    SWITCH_RANGE(b, 0xa0, 0xbf)  // FixRaw
+    SWITCH_RANGE(b, 0xa0, 0xbf)  // FixRaw / fixstr
         int count = b & 0x1f;
         if(count == 0) {
             return object_complete_string(uk, rb_str_buf_new(0));
         }
         /* read_raw_body_begin sets uk->reading_raw */
         uk->reading_raw_remaining = count;
-        return read_raw_body_begin(uk);
+        return read_raw_body_begin(uk, true);
 
     SWITCH_RANGE(b, 0x90, 0x9f)  // FixArray
         int count = b & 0x0f;
@@ -321,12 +327,9 @@ static int read_primitive(msgpack_unpacker_t* uk)
         case 0xc3:  // true
             return object_complete(uk, Qtrue);
 
-        //case 0xc4:
-        //case 0xc5:
-        //case 0xc6:
-        //case 0xc7:
-        //case 0xc8:
-        //case 0xc9:
+        //case 0xc7: // ext 8
+        //case 0xc8: // ext 16
+        //case 0xc9: // ext 32
 
         case 0xca:  // float
             {
@@ -398,14 +401,25 @@ static int read_primitive(msgpack_unpacker_t* uk)
                 return object_complete(uk, rb_ll2inum(i64));
             }
 
-        //case 0xd4:
-        //case 0xd5:
-        //case 0xd6:  // big integer 16
-        //case 0xd7:  // big integer 32
-        //case 0xd8:  // big float 16
-        //case 0xd9:  // big float 32
+        //case 0xd4:  // fixext 1
+        //case 0xd5:  // fixext 2
+        //case 0xd6:  // fixext 4
+        //case 0xd7:  // fixext 8
+        //case 0xd8:  // fixext 16
 
-        case 0xda:  // raw 16
+        case 0xd9:  // raw 8 / str 8
+            {
+                READ_CAST_BLOCK_OR_RETURN_EOF(cb, uk, 1);
+                uint8_t count = cb->u8;
+                if(count == 0) {
+                    return object_complete_string(uk, rb_str_buf_new(0));
+                }
+                /* read_raw_body_begin sets uk->reading_raw */
+                uk->reading_raw_remaining = count;
+                return read_raw_body_begin(uk, true);
+            }
+
+        case 0xda:  // raw 16 / str 16
             {
                 READ_CAST_BLOCK_OR_RETURN_EOF(cb, uk, 2);
                 uint16_t count = _msgpack_be16(cb->u16);
@@ -414,10 +428,10 @@ static int read_primitive(msgpack_unpacker_t* uk)
                 }
                 /* read_raw_body_begin sets uk->reading_raw */
                 uk->reading_raw_remaining = count;
-                return read_raw_body_begin(uk);
+                return read_raw_body_begin(uk, true);
             }
 
-        case 0xdb:  // raw 32
+        case 0xdb:  // raw 32 / str 32
             {
                 READ_CAST_BLOCK_OR_RETURN_EOF(cb, uk, 4);
                 uint32_t count = _msgpack_be32(cb->u32);
@@ -426,7 +440,43 @@ static int read_primitive(msgpack_unpacker_t* uk)
                 }
                 /* read_raw_body_begin sets uk->reading_raw */
                 uk->reading_raw_remaining = count;
-                return read_raw_body_begin(uk);
+                return read_raw_body_begin(uk, true);
+            }
+
+        case 0xc4:  // bin 8
+            {
+                READ_CAST_BLOCK_OR_RETURN_EOF(cb, uk, 1);
+                uint8_t count = cb->u8;
+                if(count == 0) {
+                    return object_complete_binary(uk, rb_str_buf_new(0));
+                }
+                /* read_raw_body_begin sets uk->reading_raw */
+                uk->reading_raw_remaining = count;
+                return read_raw_body_begin(uk, false);
+            }
+
+        case 0xc5:  // bin 16
+            {
+                READ_CAST_BLOCK_OR_RETURN_EOF(cb, uk, 2);
+                uint16_t count = _msgpack_be16(cb->u16);
+                if(count == 0) {
+                    return object_complete_binary(uk, rb_str_buf_new(0));
+                }
+                /* read_raw_body_begin sets uk->reading_raw */
+                uk->reading_raw_remaining = count;
+                return read_raw_body_begin(uk, false);
+            }
+
+        case 0xc6:  // bin 32
+            {
+                READ_CAST_BLOCK_OR_RETURN_EOF(cb, uk, 4);
+                uint32_t count = _msgpack_be32(cb->u32);
+                if(count == 0) {
+                    return object_complete_binary(uk, rb_str_buf_new(0));
+                }
+                /* read_raw_body_begin sets uk->reading_raw */
+                uk->reading_raw_remaining = count;
+                return read_raw_body_begin(uk, false);
             }
 
         case 0xdc:  // array 16
@@ -486,7 +536,7 @@ int msgpack_unpacker_read_array_header(msgpack_unpacker_t* uk, uint32_t* result_
         return b;
     }
 
-    if(0x90 < b && b < 0x9f) {
+    if(0x90 <= b && b <= 0x9f) {
         *result_size = b & 0x0f;
 
     } else if(b == 0xdc) {
@@ -503,6 +553,7 @@ int msgpack_unpacker_read_array_header(msgpack_unpacker_t* uk, uint32_t* result_
         return PRIMITIVE_UNEXPECTED_TYPE;
     }
 
+    reset_head_byte(uk);
     return 0;
 }
 
@@ -513,7 +564,7 @@ int msgpack_unpacker_read_map_header(msgpack_unpacker_t* uk, uint32_t* result_si
         return b;
     }
 
-    if(0x80 < b && b < 0x8f) {
+    if(0x80 <= b && b <= 0x8f) {
         *result_size = b & 0x0f;
 
     } else if(b == 0xde) {
@@ -530,6 +581,7 @@ int msgpack_unpacker_read_map_header(msgpack_unpacker_t* uk, uint32_t* result_si
         return PRIMITIVE_UNEXPECTED_TYPE;
     }
 
+    reset_head_byte(uk);
     return 0;
 }
 
@@ -561,7 +613,22 @@ int msgpack_unpacker_read(msgpack_unpacker_t* uk, size_t target_stack_depth)
                 top->type = STACK_TYPE_MAP_VALUE;
                 break;
             case STACK_TYPE_MAP_VALUE:
-                rb_hash_aset(top->object, top->key, uk->last_object);
+                if(uk->symbolize_keys && rb_type(top->key) == T_STRING) {
+                    /* here uses rb_intern_str instead of rb_intern so that Ruby VM can GC unused symbols */
+#ifdef HAVE_RB_STR_INTERN
+                    /* rb_str_intern is added since MRI 2.2.0 */
+                    rb_hash_aset(top->object, rb_str_intern(top->key), uk->last_object);
+#else
+#ifndef HAVE_RB_INTERN_STR
+                    /* MRI 1.8 doesn't have rb_intern_str or rb_intern2 */
+                    rb_hash_aset(top->object, ID2SYM(rb_intern(RSTRING_PTR(top->key))), uk->last_object);
+#else
+                    rb_hash_aset(top->object, ID2SYM(rb_intern_str(top->key)), uk->last_object);
+#endif
+#endif
+                } else {
+                    rb_hash_aset(top->object, top->key, uk->last_object);
+                }
                 top->type = STACK_TYPE_MAP_KEY;
                 break;
             }
@@ -662,8 +729,14 @@ int msgpack_unpacker_peek_next_object_type(msgpack_unpacker_t* uk)
         case 0xd3:  // signed int 64
             return TYPE_INTEGER;
 
-        case 0xda:  // raw 16
-        case 0xdb:  // raw 32
+        case 0xd9:  // raw 8 / str 8
+        case 0xda:  // raw 16 / str 16
+        case 0xdb:  // raw 32 / str 32
+            return TYPE_RAW;
+
+        case 0xc4:  // bin 8
+        case 0xc5:  // bin 16
+        case 0xc6:  // bin 32
             return TYPE_RAW;
 
         case 0xdc:  // array 16
